package hu.kozadat.search.presentation.viewer;

import hu.kozadat.search.model.IndexerSearchResult;
import hu.kozadat.search.model.IndexerSearchResultItem;
import hu.kozadat.search.model.IndexerSearchResultSchemaType;
import hu.kozadat.search.service.simplesearch.SimpleSearch;
import hu.kozadat.search.service.viewer.AttachmentViewerService;
import hu.scriptum.nda.api.MDcSearch;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Controller;
import org.springframework.web.bind.annotation.PathVariable;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RequestParam;
import org.springframework.web.servlet.ModelAndView;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.util.LinkedList;
import java.util.List;

/**
 * Created with IntelliJ IDEA.
 * User: tnagy
 * Date: 2013.04.10.
 * Time: 11:04
 * To change this template use File | Settings | File Templates.
 */
@Controller
public class AttachmentViewerController {

    public static final String MAPPING_PREFIX = "/forras/";
    public static final String RECORD_MAPPING = MAPPING_PREFIX + "{recordId}";
    public static final String STATIC_CONTENT_MAPPING = RECORD_MAPPING + "/" + "{staticContent:.+}";

    private static final Logger LOGGER = LoggerFactory.getLogger(AttachmentViewerController.class);

    private SimpleSearch simpleSearch;
    private AttachmentViewerService attachmentViewerService;

    @RequestMapping(STATIC_CONTENT_MAPPING)
    public ModelAndView showStaticContent(HttpServletRequest httpServletRequest, HttpServletResponse httpServletResponse, @PathVariable("recordId") String recordId, @PathVariable("staticContent") String staticContent) throws Exception {

        if (!attachmentViewerService.hasContent(recordId)) {
            return new ModelAndView("redirect:/forras/" + recordId);
        }

        String encoding = attachmentViewerService.getEncodingInfo(recordId);
        String contentType = attachmentViewerService.getContentTypeInfo(recordId);

        if (encoding != null) {
            httpServletResponse.setCharacterEncoding(encoding);
        }
        if (contentType != null) {
            httpServletResponse.setContentType(contentType);
        }

        return new ModelAndView("forward:/archiv/" + recordId + "/" + staticContent);
    }

    @RequestMapping(RECORD_MAPPING)
    public ModelAndView findAttachment(HttpServletRequest httpServletRequest, HttpServletResponse httpServletResponse, @PathVariable("recordId") String recordId, @RequestParam(value = "forceUpdate", required = false) String forceUpdate, @RequestParam(value = "referer", required = false) String referer) throws Exception {

        String sessionId = httpServletRequest.getSession().getId();

        // tnagy find record
        IndexerSearchResultItem result = findRecord(sessionId, recordId);

        if (result == null) {
            // todo tnagy use referer
            LOGGER.warn("findAttachment(): record not found {}", recordId);
            return new ModelAndView("redirect:/");
        }

        // tnagy check schema
        if (!IndexerSearchResultSchemaType.KOZADAT_ARCHIV.equals(result.getSchemaType())) {
            LOGGER.warn("findAttachment(): record {} is not in kozadat_archive schema", result.getRecordId());
            return new ModelAndView("redirect:/rekord/" + result.getRecordId());
        }

        // tnagy find source
        String sourceValue = findSourceValue(result);

        if (sourceValue == null || "".equals(sourceValue.trim())) {
            // todo tnagy use referer
            LOGGER.warn("findAttachment(): source is null or empty in record {}", result.getRecordId());
            return new ModelAndView("redirect:/rekord/" + result.getRecordId());
        }

        String guessedIndexFile = attachmentViewerService.getIndexInfo(result.getRecordId());

        //tnagy check if exists or has to be updated
        boolean downloadAndUnzip = false;
        if (forceUpdate != null && !"".equals(forceUpdate.trim())) {
            downloadAndUnzip = true;
        }
        if (!downloadAndUnzip && !attachmentViewerService.hasContent(result.getRecordId())) {
            downloadAndUnzip = true;
        }
        if (guessedIndexFile == null || "".equals(guessedIndexFile.trim())) {
            downloadAndUnzip = true;
        }

        if (downloadAndUnzip) {

            // tnagy check if it is a zipped wget result
            if (!attachmentViewerService.isZip(sourceValue)) {
                // tnagy redirect to download
                LOGGER.warn("findAttachment(): record {}'s attachment is not zip, redirect to webfarm", result.getRecordId());
                return new ModelAndView("redirect:" + sourceValue);
            }

            // tnagy download, extract and show
            guessedIndexFile = attachmentViewerService.downloadAndUnzip(result.getRecordId(), sourceValue);

        } else {

            // tnagy read index file name and forward
            guessedIndexFile = attachmentViewerService.getIndexInfo(result.getRecordId());

        }

        if (guessedIndexFile == null || "".equals(guessedIndexFile.trim())) {
            LOGGER.warn("findAttachment(): cannot guess index file name for record {}", result.getRecordId());
            return new ModelAndView("redirect:/rekord/" + result.getRecordId());
        }

        StringBuilder archiveContentUrl = new StringBuilder("redirect:/forras/");
        archiveContentUrl.append(result.getRecordId()).append("/");
        archiveContentUrl.append(guessedIndexFile);

        return new ModelAndView(archiveContentUrl.toString());
    }

    private String findSourceValue(IndexerSearchResultItem result) {
        if (result != null) {
            List<String> source = result.getSchemaElementValue("source", "");
            return (source != null && !source.isEmpty()) ? source.get(0) : null;
        }
        return null;
    }

    private IndexerSearchResultItem findRecord(String sessionId, String recordId) {
        StringBuilder searchExpressionBuilder = new StringBuilder();

        searchExpressionBuilder.append(MDcSearch.FIELD_RECORD_ID).append(":");
        searchExpressionBuilder.append(MDcSearch.ID_PREFIX).append(recordId).append(MDcSearch.ID_POSTFIX);

        String searchExpression = searchExpressionBuilder.toString();

        IndexerSearchResult indexerSearchResult = simpleSearch.search(searchExpression, new LinkedList<String>(), sessionId);
        int resultSize = indexerSearchResult.getResultSize();
        if (resultSize > 0) {
            simpleSearch.fillResult(indexerSearchResult, 0, resultSize);
            simpleSearch.close(sessionId);
        }

        return indexerSearchResult.getItemsByOaiRecordId().get(recordId);
    }

    @Autowired
    public void setSimpleSearch(SimpleSearch simpleSearch) {
        this.simpleSearch = simpleSearch;
    }

    @Autowired
    public void setAttachmentViewerService(AttachmentViewerService attachmentViewerService) {
        this.attachmentViewerService = attachmentViewerService;
    }
}
