package hu.kozadat.search.service.export;

import com.actiwise.einfoszab.Utils;
import hu.kozadat.search.exception.CSVExportException;
import hu.kozadat.search.model.IndexerSearchResult;
import hu.kozadat.search.model.IndexerSearchResultItem;
import hu.kozadat.search.persistence.SearchCache;
import org.apache.commons.csv.CSVFormat;
import org.apache.commons.csv.CSVPrinter;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.charset.Charset;
import java.util.*;

/**
 * Created by tnagy on 2016.08.17..
 */
@Service
public class ExportService {

    private static final Logger LOGGER = LoggerFactory.getLogger(ExportService.class);
    private static final String EMPTY_VALUE = "";
    private static final String CSV_COLUMN_KEY_SCHEMA_NAME = "schemaName";
    private static final String CSV_COLUMN_KEY_RECORD_ID = "recordId";
    private static final String CSV_COLUMN_KEY_DATA_PROVIDER_NAME = "dataProviderName";
    private static final String CSV_COLUMN_KEY_DATA_PROVIDER_ID = "dataProviderId";
    private static final String CSV_COLUMN_LABEL_DAVAPROVIDER_ID = "Adatgazda azonos\u00EDt\u00F3";
    private static final String CSV_COLUMN_LABEL_DATAPROVIDER_NAME = "Adatgazda n\u00E9v";
    private static final String CSV_COLUMN_LABEL_RECORD_ID = "Rekord azonos\u00EDt\u00F3";
    private static final String CSV_COLUMN_LABEL_SCHEMA_NAME = "S\u00E9ma";
    private static final Charset CHARSET_LATIN2 = Charset.forName("ISO-8859-2");
    private static final Charset CHARSET_UTF_8 = Charset.forName("UTF-8");

    private SearchCache searchCache;

    public CSVExportResult exportToCSV(IndexerSearchResult indexerSearchResult) throws CSVExportException {

        if (indexerSearchResult == null) {
            return new CSVExportResult();
        }

        long start = System.currentTimeMillis();

        if (LOGGER.isDebugEnabled()) {
            LOGGER.debug("exportToCSV(): Exporting search result as CSV...");
        }

        /**
         * Structure for rows
         */
        Map<String, Map<String, Map<String, List<String>>>> rows = new TreeMap<>();
        /**
         * Structure for columns
         */
        Map<String, SchemaElement> columnOccurrences = new LinkedHashMap<>();

        // tnagy process search result
        for (IndexerSearchResultItem item : indexerSearchResult.getItems()) {
            Long dataProviderId = item.getDataProviderId();
            String dataProviderName = item.getDataProviderName();
            String recordId = item.getRecordId();
            String schemaName = item.getSchemaName();

            // tnagy add an empty row
            if (!rows.containsKey(recordId)) {
                rows.put(recordId, new LinkedHashMap<String, Map<String, List<String>>>());
            }
            Map<String, Map<String, List<String>>> row = rows.get(recordId);

            // tnagy add metadata column values
            row.put(CSV_COLUMN_KEY_DATA_PROVIDER_ID, getValueMap(Long.toString(dataProviderId)));
            row.put(CSV_COLUMN_KEY_DATA_PROVIDER_NAME, getValueMap(dataProviderName));
            row.put(CSV_COLUMN_KEY_RECORD_ID, getValueMap(recordId));
            row.put(CSV_COLUMN_KEY_SCHEMA_NAME, getValueMap(schemaName));

            // tnagy process schema elements
            Map<String, Map<String, List<String>>> schemaElements = item.getSchemaElements();
            for (String element : schemaElements.keySet()) {
                // tnagy add element to row
                if (!row.containsKey(element)) {
                    row.put(element, new TreeMap<String, List<String>>());
                }
                // tnagy process element qualifiers
                for (String qualifier : schemaElements.get(element).keySet()) {
                    List<String> values = new LinkedList<>();
                    // tnagy remove highlight markup from values
                    List<String> valueList = schemaElements.get(element).get(qualifier);
                    for (String aValue : valueList) {
                        values.add(Utils.removeMarkup(aValue));
                    }
                    // tnagy add initial qualifier to row element
                    if (!row.get(element).containsKey(qualifier)) {
                        row.get(element).put(qualifier, new LinkedList<String>());
                    }
                    // tnagy add values to element/qualifier entry
                    row.get(element).get(qualifier).addAll(values);
                    // tnagy update occurrence to the largest value
                    String elementKey = SchemaElement.getKey(element, qualifier);
                    if (!columnOccurrences.containsKey(elementKey)) {
                        columnOccurrences.put(elementKey, new SchemaElement(element, qualifier, values.size()));
                    } else {
                        columnOccurrences.put(elementKey, new SchemaElement(element, qualifier, Math.max(columnOccurrences.get(elementKey).getOccurrence(), values.size())));
                    }
                }
            }
        }

        if (LOGGER.isDebugEnabled()) {
            LOGGER.debug("exportToCSV(): columns={}", columnOccurrences);
            LOGGER.debug("exportToCSV(): rows={}", rows);
        }

        // tnagy add all elements to all rows with empty value if necessary
        for (String recordId : rows.keySet()) {
            Map<String, Map<String, List<String>>> row = rows.get(recordId);
            for (String elementKey : columnOccurrences.keySet()) {
                SchemaElement schemaElement = columnOccurrences.get(elementKey);
                if (!row.containsKey(schemaElement.getElement())) {
                    row.put(schemaElement.getElement(), new LinkedHashMap<String, List<String>>());
                }
                if (!row.get(schemaElement.getElement()).containsKey(schemaElement.getQualifier())) {
                    List<String> values = Collections.nCopies(schemaElement.getOccurrence(), EMPTY_VALUE);
                    row.get(schemaElement.getElement()).put(schemaElement.getQualifier(), new LinkedList<>(values));
                } else {
                    List<String> values = row.get(schemaElement.getElement()).get(schemaElement.getQualifier());
                    while (values.size() < schemaElement.getOccurrence()) {
                        values.add(EMPTY_VALUE);
                    }
                }
            }
        }

        if (LOGGER.isDebugEnabled()) {
            LOGGER.debug("exportToCSV(): Table created in {} ms.", (System.currentTimeMillis() - start));
        }

        long csvPrintStart = System.currentTimeMillis();

        // tnagy create headers
        List<String> headers = new LinkedList<>();
        // tnagy add metadata header labels
        headers.add(CSV_COLUMN_LABEL_DAVAPROVIDER_ID);
        headers.add(CSV_COLUMN_LABEL_DATAPROVIDER_NAME);
        headers.add(CSV_COLUMN_LABEL_RECORD_ID);
        headers.add(CSV_COLUMN_LABEL_SCHEMA_NAME);
        // tnagy add header labels
        for (String elementKey : columnOccurrences.keySet()) {
            SchemaElement schemaElement = columnOccurrences.get(elementKey);
            String localizedElementAndQualifierName;
            if ("".equals(schemaElement.getQualifier())) {
                localizedElementAndQualifierName = searchCache.getHungarianElementName(schemaElement.getElement());
            } else {
                localizedElementAndQualifierName = searchCache.getHungarianQualifierName(schemaElement.getQualifier());
            }
            List<String> header = Collections.nCopies(schemaElement.getOccurrence(), localizedElementAndQualifierName);
            if (header.size() > 1) {
                List<String> indexedHeader = new LinkedList<>();
                for (int i = 0; i < header.size(); i++) {
                    String aHeader;
                    if (i > 0) {
                        aHeader = header.get(i) + "_" + i;
                    } else {
                        aHeader = header.get(i);
                    }
                    indexedHeader.add(aHeader);
                }
                header = indexedHeader;
            }
            headers.addAll(header);
        }

        if (LOGGER.isDebugEnabled()) {
            LOGGER.debug("exportToCSV(): headers={}", headers);
        }

        // tnagy assemble CSV
        final Appendable csvResult = new StringBuilder();
        try {

            CSVPrinter csvPrinter = CSVFormat.DEFAULT.withHeader(headers.toArray(new String[headers.size()])).print(csvResult);

            for (String recordId : rows.keySet()) {
                Map<String, Map<String, List<String>>> row = rows.get(recordId);
                List<String> values = new LinkedList<>();
                // tnagy add metadata column values
                values.addAll(row.get(CSV_COLUMN_KEY_DATA_PROVIDER_ID).get(""));
                values.addAll(row.get(CSV_COLUMN_KEY_DATA_PROVIDER_NAME).get(""));
                values.addAll(row.get(CSV_COLUMN_KEY_RECORD_ID).get(""));
                values.addAll(row.get(CSV_COLUMN_KEY_SCHEMA_NAME).get(""));
                // tnagy add record column values
                for (String elementKey : columnOccurrences.keySet()) {
                    SchemaElement schemaElement = columnOccurrences.get(elementKey);
                    values.addAll(row.get(schemaElement.getElement()).get(schemaElement.getQualifier()));
                }
                csvPrinter.printRecord(values);
            }

        } catch (IOException e) {
            throw new CSVExportException(e.getMessage(), e);
        }

        if (LOGGER.isDebugEnabled()) {
            LOGGER.debug("exportToCSV(): CSV created in {} ms.", (System.currentTimeMillis() - csvPrintStart));
        }

        // tnagy create CSV string
        String csvResultString = csvResult.toString();

        if (LOGGER.isDebugEnabled()) {
            LOGGER.debug("exportToCSV(): CSV result={}", csvResultString);
        }

        // tnagy create and return result object
        CSVExportResult result = new CSVExportResult();
        // tnagy convert UTF-8 string result to ISO-8859-2 byte array
        result.setContent(CHARSET_LATIN2.encode(CHARSET_UTF_8.decode(ByteBuffer.wrap(csvResultString.getBytes()))).array());
        result.setCharset(CHARSET_LATIN2);
        return result;
    }

    private TreeMap<String, List<String>> getValueMap(String value) {
        TreeMap<String, List<String>> valueMap = new TreeMap<>();
        valueMap.put("", Collections.singletonList(value));
        return valueMap;
    }

    @Autowired
    public void setSearchCache(SearchCache searchCache) {
        this.searchCache = searchCache;
    }

    private static class SchemaElement {
        private String schema;
        private String element;
        private String qualifier;
        private int occurrence = 1;

        public SchemaElement(String schema, String element, String qualifier, int occurrence) {
            this.schema = schema;
            this.element = element;
            this.qualifier = qualifier;
            this.occurrence = occurrence;
        }

        public SchemaElement(String element, String qualifier, int occurrence) {
            this.element = element;
            this.qualifier = qualifier;
            this.occurrence = occurrence;
        }

        public static String getKey(String element, String qualifier) {
            if ("".equals(qualifier)) {
                return element;
            }
            return qualifier;
        }

        public String getSchema() {
            return schema;
        }

        public String getElement() {
            return element;
        }

        public String getQualifier() {
            return qualifier;
        }

        public int getOccurrence() {
            return occurrence;
        }

        @Override
        public String toString() {
            return "SchemaElement{" +
                    "element='" + element + '\'' +
                    ", schema='" + schema + '\'' +
                    ", qualifier='" + qualifier + '\'' +
                    ", occurrence=" + occurrence +
                    '}';
        }
    }

}
