package hu.kozadat.search.service.log;

import com.actiwise.einfoszab.Utils;
import hu.kozadat.search.exception.LogCentralClientServiceException;
import hu.kozadat.search.exception.SearchLoggerException;
import hu.kozadat.search.interactive.InteractiveSearchExpression;
import hu.kozadat.search.model.DataProviderDisplayInfo;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.InitializingBean;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.dao.IncorrectResultSizeDataAccessException;
import org.springframework.jdbc.core.JdbcTemplate;
import org.springframework.stereotype.Service;

import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.ResultSet;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Date;

@Service("searchLogger")
public class SearchLogger implements InitializingBean {

    private static final Logger LOGGER = LoggerFactory.getLogger(SearchLogger.class);

    private static final String LOGCENTRAL_CLIENT_MODULE_NAME_SEARCHUI = "searchui";
    private static final String LOGCENTRAL_CLIENT_EVENT_TYPE_PUBLIC = "Public";

    private static final int COL_SEARCH_EXPR_MAX_LENGTH = 400;
    private int searchExpressionMaxLength = -1;

    private JdbcTemplate jdbcTemplate;
    private LogCentralClientService logCentralClientService;

    private static final SimpleDateFormat YYYYMMDDHHMMSS_FORMATTER = new SimpleDateFormat("yyyyMMddHHmmss");
    private Date date;

    public SearchLogger() {
        LOGGER.info("Creating SearchLogger...");
        LOGGER.info("SearchLogger created.");
    }

    private static String formatDate(Date date) {
        synchronized (YYYYMMDDHHMMSS_FORMATTER) {
            return YYYYMMDDHHMMSS_FORMATTER.format(date);
        }
    }

    public static Date parseDate(String date) {
        try {
            synchronized (YYYYMMDDHHMMSS_FORMATTER) {
                return YYYYMMDDHHMMSS_FORMATTER.parse(date);
            }
        } catch (ParseException e) {
            throw new RuntimeException("invalid date");
        }
    }

    protected void logMessage(String event, String user, String dataProvider, Date time, String details) throws SearchLoggerException {
        try {
            logCentralClientService.sendMessage(LOGCENTRAL_CLIENT_MODULE_NAME_SEARCHUI, LOGCENTRAL_CLIENT_EVENT_TYPE_PUBLIC, event, user, dataProvider, time, details);
        } catch (LogCentralClientServiceException e) {
            throw new SearchLoggerException(e.getMessage(), e);
        }
    }

    public void addSimpleSearchLog(String clientIp, String query, String schemaId, int fullElapsedTime, int resultSize) throws SearchLoggerException {
        Date date = new Date();
        jdbcTemplate.update("insert into search_logs(client_ipaddress, start_time, " +
                "elapsed_time, search_type, search_expr, selected_schema_id, result_size) " +
                "values (?, ?, ?, ?, ?, ?, ?)",
                new Object[]{
                        clientIp,
                        formatDate(date),
                        new Integer(fullElapsedTime),
                        "S",
                        Utils.truncate(Utils.removeMarkup(query), searchExpressionMaxLength),
                        schemaId != null ? ("__all__".equals(schemaId) ? null : new Integer(schemaId)) : null,
                        new Integer(resultSize)
                });
        StringBuilder detailsBuilder = new StringBuilder();
        detailsBuilder.append("Kereső kifejezés: ").append(query).append(". ");
        detailsBuilder.append("Találatok: ").append(resultSize).append(".");

        logMessage("SIMPLE_SEARCH", "", "", date, detailsBuilder.toString());
    }

    public void addInstitutionSearchLog(String clientIp, String institutionName, Long institutionId, int fullElapsedTime, int resultSize) throws SearchLoggerException {
        Date date = new Date();
        jdbcTemplate.update("insert into search_logs(client_ipaddress, start_time, " +
                "elapsed_time, search_type, search_expr, selected_dp_id, result_size) " +
                "values (?, ?, ?, ?, ?, ?, ?)",
                new Object[]{
                        clientIp,
                        formatDate(date),
                        new Integer(fullElapsedTime),
                        "I",
                        Utils.truncate(Utils.removeMarkup(institutionName), searchExpressionMaxLength),
                        institutionId,
                        new Integer(resultSize)
                });
        StringBuilder detailsBuilder = new StringBuilder();
        detailsBuilder.append("Kereső kifejezés: ").append(institutionName).append(". ");
        detailsBuilder.append("Találatok: ").append(resultSize).append(".");

        if (institutionId == null) {
            logMessage("INSTITUTION_SEARCH", "", "", date, detailsBuilder.toString());
        } else {
            String oaiId = (String) jdbcTemplate.queryForObject("select oai_id from data_providers where id = ?", new Object[]{institutionId}, String.class);
            logMessage("INSTITUTION_RECORDS_VIEW", "", oaiId, date, detailsBuilder.toString());
        }
    }

    public void addInteractiveSearchLog(InteractiveSearchExpression expr) throws SearchLoggerException {
        Date date = new Date();
        logMessage("INTERACTIVE_SEARCH", "", "", date, expr.toString());
    }

    public void addInstitutionInfo(DataProviderDisplayInfo dpi) throws SearchLoggerException {
        Date date = new Date();
        logMessage("INSTITUTIONINFO", "", dpi.getOaiId() == null ? "" : dpi.getOaiId(), date, "");
    }

    public void addMetarecordViewLog(String clientIp, String metarecordId, long dataProviderId, String title) throws SearchLoggerException {
        String oaiId = null;
        try {
            oaiId = (String) jdbcTemplate.queryForObject("SELECT oai_id FROM xmls WHERE id=" + metarecordId, String.class);
        } catch (IncorrectResultSizeDataAccessException e) {
            LOGGER.warn(e.getMessage(), e);
        }

        date = new Date();
        jdbcTemplate.update("insert into search_logs(client_ipaddress, start_time, " +
                "search_type, search_expr, selected_dp_id, oai_id) values (?, ?, ?, ?, ?, ?)",
                new Object[]{
                        clientIp,
                        formatDate(date),
                        "M",
                        Utils.truncate(Utils.removeMarkup(title), searchExpressionMaxLength),
                        new Long(dataProviderId),
                        oaiId
                });
        StringBuilder detailsBuilder = new StringBuilder();
        detailsBuilder.append("Kereső kifejezés: ").append(title).append(". ");
        detailsBuilder.append("OAI azonosító: ").append(oaiId).append(". ");
        detailsBuilder.append("Metarekord azonosító: ").append(metarecordId).append(".");
        logMessage("METARECORD_VIEW", "", "", date, detailsBuilder.toString());
    }

    public void addOriginalViewLog(String clientIp, String metarecordId, long dataProviderId, String title) throws SearchLoggerException {
        String oaiId = null;
        try {
            oaiId = (String) jdbcTemplate.queryForObject("SELECT oai_id FROM xmls WHERE id=" + metarecordId, String.class);
        } catch (IncorrectResultSizeDataAccessException e) {
            LOGGER.warn(e.getMessage(), e);
        }

        jdbcTemplate.update("insert into search_logs(client_ipaddress, start_time, " +
                "search_type, search_expr, selected_dp_id, oai_id) values (?, ?, ?, ?, ?, ?)",
                new Object[]{
                        clientIp,
                        formatDate(new Date()),
                        "R",
                        Utils.truncate(Utils.removeMarkup(title), searchExpressionMaxLength),
                        new Long(dataProviderId),
                        oaiId
                });
        StringBuilder detailsBuilder = new StringBuilder();
        detailsBuilder.append("Kereső kifejezés: ").append(title).append(". ");
        detailsBuilder.append("OAI azonosító: ").append(oaiId).append(". ");
        detailsBuilder.append("Metarekord azonosító: ").append(metarecordId).append(".");
        logMessage("RECORD_VIEW", "", "", date, detailsBuilder.toString());
    }

    @Autowired
    public void setJdbcTemplate(JdbcTemplate jdbcTemplate) {
        this.jdbcTemplate = jdbcTemplate;
    }

    @Autowired
    public void setLogCentralClientService(LogCentralClientService logCentralClientService) {
        this.logCentralClientService = logCentralClientService;
    }

    public String toString() {
        return "SearchLogger version: $Revision$ ($Date$)";
    }

    // Log-oljuk az adott megvalositas verziojat
    static {
        LOGGER.info("[CLASSVERSION] $Revision$ ($Date$)");
    }

    @Override
    public void afterPropertiesSet() throws Exception {
        findExpressionMaxLength();
        if (searchExpressionMaxLength == -1) {
            searchExpressionMaxLength = COL_SEARCH_EXPR_MAX_LENGTH;
        }
    }

    private void findExpressionMaxLength() {
        Connection connection = null;
        try {
            connection = jdbcTemplate.getDataSource().getConnection();
            DatabaseMetaData metaData = connection.getMetaData();
            ResultSet columns = metaData.getColumns(null, null, "SEARCH_LOGS", "SEARCH_EXPR");
            while (columns.next()) {
                searchExpressionMaxLength = columns.getInt("COLUMN_SIZE");
            }
            if (searchExpressionMaxLength == -1) {
                LOGGER.info("findExpressionMaxLength(): SEARCH_EXPR size not found.");
            } else {
                LOGGER.info("findExpressionMaxLength(): SEARCH_EXPR size found: {}", searchExpressionMaxLength);
            }
        } catch (Exception e) {
            LOGGER.error(e.getMessage(), e);
        } finally {
            if (connection != null) {
                try {
                    connection.close();
                } catch (Exception e) {
                    LOGGER.error(e.getMessage(), e);
                }
            }
        }
    }
}
