package hu.kozadat.search.service.simplesearch;

import hu.kozadat.search.base.preprocessor.QueryPreprocessorManager;
import hu.kozadat.search.base.preprocessor.QueryPreprocessorResult;
import hu.kozadat.search.model.IndexerSearchResult;
import hu.kozadat.search.model.IndexerSearchResultItem;
import hu.kozadat.search.service.IndexerSearchResultParser;
import hu.kozadat.search.service.SearchUIConfiguration;
import hu.scriptum.nda.api.MDcSearch;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.InitializingBean;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import java.util.LinkedList;
import java.util.List;

@Service("simpleSearch")
public class SimpleSearch implements InitializingBean {

    private static final Logger LOGGER = LoggerFactory.getLogger(SimpleSearch.class);

    private SearchUIConfiguration searchUIConfiguration;
    private MDcSearch searchService;
    private int maxResultCount = 10000;
    private QueryPreprocessorManager queryPreprocessor;

    public IndexerSearchResult search(String searchExpression, List<String> schemaIds, String sessionId) {
        IndexerSearchResult indexerSearchResult = new IndexerSearchResult();
        indexerSearchResult.setSessionId(sessionId);

        QueryPreprocessorResult preprocessedQuery = queryPreprocessor.preprocess(searchExpression != null ? searchExpression.toLowerCase() : null, true);
        String searchQuery = preprocessedQuery.getFilteredQuery();

        if (schemaIds != null && !schemaIds.isEmpty()) {
            StringBuilder searchQueryBuilder = new StringBuilder();
            searchQueryBuilder.append("(");
            searchQueryBuilder.append(searchQuery);
            searchQueryBuilder.append(")");
            searchQueryBuilder.append(" AND ");
            searchQueryBuilder.append("(");
            for (int i = 0; i < schemaIds.size(); i++) {
                String schemaId = schemaIds.get(i);
                if (i > 0) {
                    searchQueryBuilder.append(" OR ");
                }
                searchQueryBuilder.append(MDcSearch.FIELD_SCHEMA_ID);
                searchQueryBuilder.append(":");
                searchQueryBuilder.append(MDcSearch.ID_PREFIX);
                searchQueryBuilder.append(schemaId);
                searchQueryBuilder.append(MDcSearch.ID_POSTFIX);
            }
            searchQueryBuilder.append(")");

            searchQuery = searchQueryBuilder.toString();
        }

        indexerSearchResult.setQuery(searchQuery);
        try {
            long time = System.currentTimeMillis();
            int resultNum = 0;
            try {
                resultNum = searchService.executeQuery(sessionId, searchQuery);
            } catch (Exception e) {
                LOGGER.error("Simple search failed: " + e.getMessage(), e);
            }

            long value = System.currentTimeMillis() - time;
            indexerSearchResult.setResultSize(resultNum);
            indexerSearchResult.setExecutionTime(value);

        } catch (Exception ex) {
            LOGGER.error(ex.getMessage(), ex);
        }
        indexerSearchResult.setCharactersRemovedFromQuery(preprocessedQuery.getRemovedCharsAsString());

        return indexerSearchResult;
    }

    public IndexerSearchResult search(String searchExpression, String schemaId, String sessionId) {
        List<String> schemaIds = new LinkedList<String>();
        if (!"__all__".equals(schemaId)) {
            schemaIds.add(schemaId);
        }
        return search(searchExpression, schemaIds, sessionId);
    }

    public IndexerSearchResult fillResult(IndexerSearchResult result, int fromIndex, int numOfHits) {

        IndexerSearchResultParser indexerSearchResultParser = new IndexerSearchResultParser();

        try {
            long time = System.currentTimeMillis();

            String[] hits = searchService.getHits(result.getSessionId(), fromIndex, Math.max(numOfHits, maxResultCount));

            for (String hit : hits) {

                IndexerSearchResultItem indexerSearchResultItem = indexerSearchResultParser.parse(hit);
                result.addItem(indexerSearchResultItem);

            }

            long resultProcessTime = System.currentTimeMillis() - time;
            result.setProcessTime(resultProcessTime);

            return result;


        } catch (Exception e) {
            LOGGER.error(e.getMessage(), e);
            return result;
        }

    }

    public IndexerSearchResult searchInstitution(Long dataProviderId, String text, List<String> schemaIds, String sessionId) {

        StringBuilder searchExpressionBuilder = new StringBuilder();

        if (text != null && !"".equals(text)) {
            QueryPreprocessorResult preprocessedQuery = queryPreprocessor.preprocess(text.toLowerCase(), true);
            String searchQuery = preprocessedQuery.getFilteredQuery();

            searchExpressionBuilder.append("(");
            searchExpressionBuilder.append(searchQuery);
            searchExpressionBuilder.append(")");
        }

        if (dataProviderId != null) {
            if (searchExpressionBuilder.length() > 0) {
                searchExpressionBuilder.append(" AND ");
            }
            searchExpressionBuilder.append("(");
            searchExpressionBuilder.append(MDcSearch.FIELD_PROVIDER_ID);
            searchExpressionBuilder.append(":");
            searchExpressionBuilder.append(MDcSearch.ID_PREFIX);
            searchExpressionBuilder.append(dataProviderId);
            searchExpressionBuilder.append(MDcSearch.ID_POSTFIX);
            searchExpressionBuilder.append(")");
        }

        if (schemaIds != null && !schemaIds.isEmpty()) {
            if (searchExpressionBuilder.length() > 0) {
                searchExpressionBuilder.append(" AND ");
            }
            searchExpressionBuilder.append("(");
            for (int i = 0; i < schemaIds.size(); i++) {
                String schemaId = schemaIds.get(i);
                if (i > 0) {
                    searchExpressionBuilder.append(" OR ");
                }
                searchExpressionBuilder.append(MDcSearch.FIELD_SCHEMA_ID);
                searchExpressionBuilder.append(":");
                searchExpressionBuilder.append(MDcSearch.ID_PREFIX);
                searchExpressionBuilder.append(schemaId);
                searchExpressionBuilder.append(MDcSearch.ID_POSTFIX);
            }
            searchExpressionBuilder.append(")");
        }

        String searchExpression = searchExpressionBuilder.toString();

        IndexerSearchResult indexerSearchResult = new IndexerSearchResult();

        indexerSearchResult.setSessionId(sessionId);
        indexerSearchResult.setQuery(searchExpression);

        long queryStart = System.currentTimeMillis();
        int resultSize = searchService.executeQuery(sessionId, searchExpression);
        long queryElapsedTime = System.currentTimeMillis() - queryStart;

        indexerSearchResult.setResultSize(resultSize);
        indexerSearchResult.setExecutionTime(queryElapsedTime);

        IndexerSearchResultParser indexerSearchResultParser = new IndexerSearchResultParser();

        long processStart = System.currentTimeMillis();

        if (resultSize > 0) {
            String[] hits = searchService.getHits(indexerSearchResult.getSessionId(), 0, Math.max(resultSize, maxResultCount));

            for (String hit : hits) {

                IndexerSearchResultItem indexerSearchResultItem = indexerSearchResultParser.parse(hit);
                indexerSearchResult.addItem(indexerSearchResultItem);

            }
        }

        long resultProcessTime = System.currentTimeMillis() - processStart;
        indexerSearchResult.setProcessTime(resultProcessTime);

        return indexerSearchResult;
    }

    public void close(String sessionId) {
        searchService.closeSession(sessionId);
    }

    @Autowired
    public void setQueryPreprocessor(QueryPreprocessorManager queryPreprocessor) {
        this.queryPreprocessor = queryPreprocessor;
    }

    @Autowired
    public void setSearchService(MDcSearch searchService) {
        this.searchService = searchService;
    }

    @Autowired
    public void setSearchUIConfiguration(SearchUIConfiguration searchUIConfiguration) {
        this.searchUIConfiguration = searchUIConfiguration;
    }

    public void afterPropertiesSet() throws Exception {
        this.maxResultCount = searchUIConfiguration.getSimpleSearchMaxResultCount();
    }
}
